import re

def test_header_barcode_extraction(fastq_test_header: str, header_barcode_regex: str):
    # Use re.search to find the barcode match in the header
    barcode_match = re.search(header_barcode_regex, fastq_test_header)

    # Check if a barcode match is found
    if barcode_match:
        # Extract the barcode from the match
        barcode = barcode_match.group(1)
        print("Barcode:", barcode)
        return barcode
    else:
        print("No barcode match found.")


def test_header_umi_extraction(fastq_test_header: str, header_umi_regex: str):
    # Use re.search to find the umi match in the header
    umi_match = re.search(header_umi_regex, fastq_test_header)

    # Check if a umi match is found
    if umi_match:
        # Extract the umi from the match
        umi = umi_match.group(1)
        print("UMI:", umi)
        return umi
    else:
        print("No UMI match found.")


from typeguard import typechecked
from ..models.error_models import GuideCountErrorType, GuideCountError
from ..models.quality_control_models import SingleInferenceQualityControlResult, MatchSetSingleInferenceQualityControlResult, QualityControlResult
from typing import DefaultDict, Tuple, Optional
from collections import defaultdict
from ..processing.crispr_editing_processing import get_non_error_dict

#
# PERFORM THE QC
#
def perform_counts_quality_control(observed_guide_reporter_umi_counts_inferred: DefaultDict[Tuple[str,Optional[str],Optional[str]], dict], contains_umi: bool, contains_surrogate: bool, contains_barcode: bool):
    get_umi_noncollapsed_counts = lambda counts_inferred_dict : sum([sum(counts_inferred_value.observed_value.values()) for counts_inferred_value in counts_inferred_dict.values()]) # NOTE: for UMI, observed_value is a Counter, so need to use .values()
    get_umi_collapsed_counts = lambda counts_inferred_dict : sum([len(counts_inferred_value.observed_value.values()) for counts_inferred_value in counts_inferred_dict.values()]) # NOTE: for UMI, observed_value is a Counter, so need to use .values()
    get_counts = lambda counts_inferred_dict : sum([counts_inferred_value.observed_value for counts_inferred_value in counts_inferred_dict.values()]) # NOTE: for UMI, observed_value is an int, so NO NEED to use .values()

    @typechecked
    def set_num_non_error_counts(single_inference_quality_control_result: SingleInferenceQualityControlResult, counts_inferred_dict: dict) -> SingleInferenceQualityControlResult:
        # TODO: Look into the MatchSetSingleInferenceMatchResultValue.matches, and calculate how many duplicates, no matches, and single matches there are. Are no_matches thrown as error?
        if contains_umi:
            single_inference_quality_control_result.num_non_error_umi_noncollapsed_counts = get_umi_noncollapsed_counts(counts_inferred_dict)
            single_inference_quality_control_result.num_non_error_umi_collapsed_counts = get_umi_collapsed_counts(counts_inferred_dict)

            single_inference_quality_control_result.num_total_umi_noncollapsed_counts = get_umi_noncollapsed_counts(observed_guide_reporter_umi_counts_inferred)
            single_inference_quality_control_result.num_total_umi_collapsed_counts = get_umi_collapsed_counts(observed_guide_reporter_umi_counts_inferred)
        else:
            single_inference_quality_control_result.num_non_error_counts = get_counts(counts_inferred_dict)
            single_inference_quality_control_result.num_total_counts = get_counts(observed_guide_reporter_umi_counts_inferred)
        return single_inference_quality_control_result

    @typechecked
    def set_num_guide_count_error_types(single_inference_quality_control_result: SingleInferenceQualityControlResult, attribute_name: str) -> SingleInferenceQualityControlResult:
        guide_count_error_type_umi_noncollapsed_count: DefaultDict[GuideCountErrorType, int] = defaultdict(int)
        guide_count_error_type_umi_collapsed_count: DefaultDict[GuideCountErrorType, int] = defaultdict(int)
        guide_count_error_type_count: DefaultDict[GuideCountErrorType, int] = defaultdict(int)
        for _, observed_guide_reporter_umi_counts_inferred_value in observed_guide_reporter_umi_counts_inferred.items():
            error_result: Optional[GuideCountError] = getattr(observed_guide_reporter_umi_counts_inferred_value.inferred_value, attribute_name).error if getattr(observed_guide_reporter_umi_counts_inferred_value.inferred_value, attribute_name) is not None else GuideCountError(guide_count_error_type=GuideCountErrorType.NULL_MATCH_RESULT)
            if error_result is not None:
                if contains_umi:
                    guide_count_error_type_umi_noncollapsed_count[error_result.guide_count_error_type] += sum(observed_guide_reporter_umi_counts_inferred_value.observed_value.values())
                    guide_count_error_type_umi_collapsed_count[error_result.guide_count_error_type] += len(observed_guide_reporter_umi_counts_inferred_value.observed_value.values())
                else:
                    guide_count_error_type_count[error_result.guide_count_error_type] += observed_guide_reporter_umi_counts_inferred_value.observed_value

        single_inference_quality_control_result.guide_count_error_type_umi_noncollapsed_count = guide_count_error_type_umi_noncollapsed_count
        single_inference_quality_control_result.guide_count_error_type_umi_collapsed_count = guide_count_error_type_umi_collapsed_count
        single_inference_quality_control_result.guide_count_error_type_count = guide_count_error_type_count

        return single_inference_quality_control_result

    @typechecked
    def set_match_set_single_inference_quality_control_results(attribute_name: str) -> MatchSetSingleInferenceQualityControlResult:
        single_inference_quality_control_result = MatchSetSingleInferenceQualityControlResult()
        single_inference_quality_control_result.non_error_dict = get_non_error_dict(observed_guide_reporter_umi_counts_inferred, attribute_name)
        
        single_inference_quality_control_result = set_num_non_error_counts(single_inference_quality_control_result, single_inference_quality_control_result.non_error_dict)
        single_inference_quality_control_result = set_num_guide_count_error_types(single_inference_quality_control_result, attribute_name)
        return single_inference_quality_control_result


    quality_control_result = QualityControlResult()
    quality_control_result.protospacer_match = set_match_set_single_inference_quality_control_results("protospacer_match")
    if contains_barcode:
        quality_control_result.protospacer_match_barcode_match = set_match_set_single_inference_quality_control_results("protospacer_match_barcode_match")
        if contains_surrogate:
            quality_control_result.protospacer_match_surrogate_match_barcode_match = set_match_set_single_inference_quality_control_results("protospacer_match_surrogate_match_barcode_match")
            quality_control_result.protospacer_mismatch_surrogate_match_barcode_match = set_match_set_single_inference_quality_control_results("protospacer_mismatch_surrogate_match_barcode_match")
    if contains_surrogate:
        quality_control_result.protospacer_match_surrogate_match = set_match_set_single_inference_quality_control_results("protospacer_match_surrogate_match")
        quality_control_result.protospacer_mismatch_surrogate_match = set_match_set_single_inference_quality_control_results("protospacer_mismatch_surrogate_match")
    
    return quality_control_result


from ..models.mapping_models import MatchSetWhitelistReporterCounterSeriesResults
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
from typing import List, Optional

def calculate_gini_index(count_series: pd.Series, eps=1e-8):
    '''
        Code from https://gist.github.com/fclesio/dfde0c512a0af55ef3c3dbdbec4bd445
    '''
    # Convert series to numpy array
    arr = count_series.to_numpy()

    # All values are treated equally, arrays must be 1d and > 0:
    arr = np.abs(arr).flatten() + eps

    # Values must be sorted:
    arr = np.sort(arr)

    # Index per array element:
    index = np.arange(1, arr.shape[0]+1)

    # Number of array elements:
    N = arr.shape[0]

    # Gini coefficient:
    return(np.sum((2*index - N - 1)*arr))/(N*np.sum(arr))

def log_count_series_warnings(match_type_string: str,
         count_type_string: str,
         zerocount: int,
         gini: float,
         std: float,
         gini_index_threshold: float,
         coverage_lower_quantile: float,
         lower_quantile_coverage_percentage: float,
         lower_quantile_coverage_threshold: float,
         coverage_mean: float,
         mean_coverage_threshold: int,
         total_count: Optional[int] = None,
         protospacer_match_gt_given_match_type_total: Optional[int] = None,
         given_match_type_vs_protospacer_match_total_ratio: Optional[float] = None,
         given_match_type_vs_protospacer_match_total_ratio_threshold: Optional[float] = None,
         umidups_mean: Optional[float] = None,
         umidups_lower_threshold: Optional[float] = None,
         umidups_upper_threshold: Optional[float] = None):
    
    if protospacer_match_gt_given_match_type_total is not None:
        # Check that protospacer-only count is greater than protospacer+(surrogate)+(barcode) count
        if protospacer_match_gt_given_match_type_total < total_count:
            print(f"WARNING for {match_type_string}; {count_type_string}: There are {total_count-protospacer_match_gt_given_match_type_total} unexpected sequences greater than the protospacer-match out of {total_count} guides. This may be due to the match type having improved mapping performance, though these sequences many be worth investigating manually.")
    if (given_match_type_vs_protospacer_match_total_ratio is not None) and (given_match_type_vs_protospacer_match_total_ratio_threshold is not None):
        # Check that the protospacer-only to protospacer+(surrogate)+(barcode) count proportion is not too low
        if given_match_type_vs_protospacer_match_total_ratio <= given_match_type_vs_protospacer_match_total_ratio_threshold:
            print(f"WARNING for {match_type_string}; {count_type_string}: The match type has a much less proportion of coverage (of {given_match_type_vs_protospacer_match_total_ratio}) compared to the protospacer-match. This may be due to high recombination or mapping issues with the given match type. Manually check if this persists across all match types and samples.")
    if zerocount > 0: 
        # Check that there are no guides with a count of zero.
        print(f"WARNING for {match_type_string}; {count_type_string}: There are {zerocount} guides with count of 0 out of {total_count} total guides. This may be due to low coverage or a mapping issue. Recommended to investigate coverage issue or investigate these sequences manually for mapping issues, especially if recurrent across different samples (especially high-coverage plasmid samples).")
    if gini >= gini_index_threshold:
        # Check that the Gini index is not too high
        print(f"WARNING for {match_type_string}; {count_type_string}: Guide counts have a high gini index of {gini} greather than threshold {gini_index_threshold} with standard deviation {std}. There may be many guides with low coverage that may be underpowered for statistical testing, even if the mean coverage is high. This may be due to a bottleneck during the screen. Recommended to compare the gini index with the plasmid sample to see how much the non-uniformity increases during the screen.")
    if coverage_mean <= mean_coverage_threshold:
        # Check that the mean coverage is not too low
        print(f"WARNING: Mean coverage {coverage_mean} is below specified threshold {mean_coverage_threshold}.")
    if coverage_lower_quantile <= lower_quantile_coverage_threshold:
        # Check that the lower quantile coverage is not too low
        print(f"WARNING for {match_type_string}; {count_type_string}: Lower {lower_quantile_coverage_percentage} quantile of coverage {coverage_lower_quantile} is below specified threshold {lower_quantile_coverage_threshold}. These guides may be underpowered for statistical testing unless the effect sizes are large.")
    
    if (umidups_mean is not None) and (umidups_lower_threshold is not None):
        # Check that the UMI duplication rate is not too low
        if umidups_mean <= umidups_lower_threshold:
            print(f"WARNING for {match_type_string}; {count_type_string}: Average UMI duplication rate of {umidups_mean} is below specified threshold of {umidups_lower_threshold}, refer to the UMI duplication histogram plotted. Your sample may be undersequenced as low average UMI duplication suggests spare material unsequenced. If guide coverage is low, it is suggested to sequence more material to increase coverage. It is also suggested to use the UMI-noncollapsed count for downstream analysis")
            
            # If the guide coverage is also low, recommend sequencing more
            if (coverage_mean <= mean_coverage_threshold) or (coverage_lower_quantile <= lower_quantile_coverage_threshold):
                print(f"WARNING for {match_type_string}; {count_type_string}: Since both the average UMI duplication rate ({umidups_mean}) and guide coverage of {coverage_mean} (with lower quantile coverage of {coverage_lower_quantile}) is low, it is suggested to sequence more material to increase guide coverage.")

    if (umidups_mean is not None) and (umidups_upper_threshold is not None):
        # Check that the UMI duplication rate is not too high.
        if umidups_mean >= umidups_upper_threshold:
            print(f"NOTE for {match_type_string}; {count_type_string}: Average UMI duplication rate of {umidups_mean} is higher than specified threshold of {umidups_upper_threshold}, refer to the UMI duplication histogram plotted. Your sample may be oversequenced as high average UMI duplication suggests material repetitively sequence post-PCR amplification. Suggested to use the UMI-collapsed count for downstream analysis.")

            # If the guide coverage is also low, recommend performing another screen with higher cell coverage to gain sufficient power
            if (coverage_mean <= mean_coverage_threshold) or (coverage_lower_quantile <= lower_quantile_coverage_threshold):
                print(f"WARNING for {match_type_string}; {count_type_string}: Since the average UMI duplication rate ({umidups_mean}) is high but the guide coverage of {coverage_mean} (with lower quantile coverage of {coverage_lower_quantile}) is low, sequencing more material may only increase UMI duplication but not the UMI-collapsed guide coverage. It is suggested to perform another screen with increased cell coverage to ensure sufficient power.")
    

def count_series_result_quality_control_stats(count_series_result: MatchSetWhitelistReporterCounterSeriesResults,
                                              contains_umi: bool, 
                                              contains_surrogate: bool, 
                                              contains_barcode: bool, 
                                              display_visualizations: bool = True,
                                              gini_index_threshold: float = 0.4,
                                              mean_coverage_threshold: float = 100,
                                              lower_quantile_coverage_percentage: float = 0.1,
                                              lower_quantile_coverage_threshold: float = 60,
                                              given_match_type_vs_protospacer_match_total_ratio_threshold= 0.3,
                                              umidups_lower_threshold: int = 2,
                                              umidups_upper_threshold: int = 4):

    
    count_series_result_quality_control_stats_dict = dict()
    if contains_umi:

        gini_index_bar_plot_labels: List[str] = []
        collapsed_gini_index_bar_plot_values: List[float] = []
        noncollapsed_gini_index_bar_plot_values: List[float] = []
        
        # PROTOSPACER - COLLAPSED
        protospacer_match_collapsed_zerocount = sum(count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries == 0)
        protospacer_match_collapsed_totalcount = len(count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries)
        protospacer_match_collapsed_gini = calculate_gini_index(count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries)
        protospacer_match_collapsed_coverage_mean = np.mean(count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries)
        protospacer_match_collapsed_coverage_std = np.std(count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries)
        protospacer_match_collapsed_coverage_lower_quantile = count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries.quantile(q=lower_quantile_coverage_percentage)
        
        count_series_result_quality_control_stats_dict.update({
            "protospacer_match_collapsed_totalcount": protospacer_match_collapsed_totalcount,
            "protospacer_match_collapsed_zerocount": protospacer_match_collapsed_zerocount,
            "protospacer_match_collapsed_gini": protospacer_match_collapsed_gini,
            "protospacer_match_collapsed_coverage_mean": protospacer_match_collapsed_coverage_mean,
            "protospacer_match_collapsed_coverage_std": protospacer_match_collapsed_coverage_std,
            f"protospacer_match_collapsed_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_match_collapsed_coverage_lower_quantile
        })

        if display_visualizations:
            # Add values for plots
            gini_index_bar_plot_labels.append("PROTOSPACER")
            collapsed_gini_index_bar_plot_values.append(protospacer_match_collapsed_gini)

            # Log warnings from count statistics
            log_count_series_warnings(match_type_string="protospacer-match",
                                      count_type_string="UMI-collapsed",
                                      total_count=None,
                                      protospacer_match_gt_given_match_type_total=None,
                                      given_match_type_vs_protospacer_match_total_ratio=None,
                                      given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                                      zerocount=protospacer_match_collapsed_zerocount,
                                      gini=protospacer_match_collapsed_gini,
                                      std=protospacer_match_collapsed_coverage_std,
                                      gini_index_threshold=gini_index_threshold,
                                      coverage_lower_quantile=protospacer_match_collapsed_coverage_lower_quantile,
                                      lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                                      lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                                      coverage_mean=protospacer_match_collapsed_coverage_mean,
                                      mean_coverage_threshold=mean_coverage_threshold)

        # PROTOSPACER - NONCOLLAPSED
        protospacer_match_noncollapsed_zerocount = sum(count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries == 0)
        protospacer_match_noncollapsed_totalcount = len(count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries)
        protospacer_match_noncollapsed_gini = calculate_gini_index(count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries)
        protospacer_match_noncollapsed_coverage_mean = np.mean(count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries)
        protospacer_match_noncollapsed_coverage_std = np.std(count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries)
        protospacer_match_noncollapsed_coverage_lower_quantile = count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries.quantile(q=lower_quantile_coverage_percentage)

        protospacer_match_umidups = count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries / count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries
        protospacer_match_umidups_mean = protospacer_match_umidups.mean()

        count_series_result_quality_control_stats_dict.update({
            "protospacer_match_noncollapsed_totalcount": protospacer_match_noncollapsed_totalcount,
            "protospacer_match_noncollapsed_zerocount": protospacer_match_noncollapsed_zerocount,
            "protospacer_match_noncollapsed_gini": protospacer_match_noncollapsed_gini,
            "protospacer_match_noncollapsed_coverage_mean": protospacer_match_noncollapsed_coverage_mean,
            "protospacer_match_noncollapsed_coverage_std": protospacer_match_noncollapsed_coverage_std,
            f"protospacer_match_noncollapsed_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_match_noncollapsed_coverage_lower_quantile,
            "protospacer_match_umidups_mean": protospacer_match_umidups_mean
        })

        if display_visualizations:
            # Add values for plots
            noncollapsed_gini_index_bar_plot_values.append(protospacer_match_noncollapsed_gini)

            # Log warnings from count statistics
            log_count_series_warnings(match_type_string="protospacer-match",
                                      count_type_string="UMI-noncollapsed",
                                      total_count=None,
                                      protospacer_match_gt_given_match_type_total=None,
                                      given_match_type_vs_protospacer_match_total_ratio=None,
                                      given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                                      zerocount=protospacer_match_noncollapsed_zerocount,
                                      gini=protospacer_match_noncollapsed_gini,
                                      std=protospacer_match_noncollapsed_coverage_std,
                                      gini_index_threshold=gini_index_threshold,
                                      coverage_lower_quantile=protospacer_match_collapsed_coverage_lower_quantile,
                                      lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                                      lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                                      coverage_mean=protospacer_match_noncollapsed_coverage_mean,
                                      mean_coverage_threshold=mean_coverage_threshold,
                                      umidups_mean=protospacer_match_umidups_mean,
                                      umidups_lower_threshold=umidups_lower_threshold,
                                      umidups_upper_threshold=umidups_upper_threshold)


        if contains_surrogate:
                # PROTOSPACER+SURROGATE vs. PROTOSPACER - COLLAPSED
                protospacer_match_gt_protospacer_surrogate_match_collapsed_total = sum(count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries >= count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_collapsed_counterseries)
                protospacer_surrogate_match_vs_protospacer_match_collapsed_total_ratio_threshold = sum(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_collapsed_counterseries) / sum(count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries)
                protospacer_surrogate_match_collapsed_zerocount = sum(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_collapsed_counterseries == 0)
                protospacer_surrogate_match_collapsed_totalcount = len(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_collapsed_counterseries)
                protospacer_surrogate_match_collapsed_gini = calculate_gini_index(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_collapsed_counterseries)
                protospacer_surrogate_match_collapsed_coverage_mean = np.mean(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_collapsed_counterseries)
                protospacer_surrogate_match_collapsed_coverage_std = np.std(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_collapsed_counterseries)
                protospacer_surrogate_match_collapsed_coverage_lower_quantile = count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_collapsed_counterseries.quantile(q=lower_quantile_coverage_percentage)

                count_series_result_quality_control_stats_dict.update({
                    "protospacer_surrogate_match_collapsed_totalcount": protospacer_surrogate_match_collapsed_totalcount,
                    "protospacer_match_gt_protospacer_surrogate_match_collapsed_total": protospacer_match_gt_protospacer_surrogate_match_collapsed_total,
                    "protospacer_surrogate_match_vs_protospacer_match_collapsed_total_ratio_threshold": protospacer_surrogate_match_vs_protospacer_match_collapsed_total_ratio_threshold,
                    "protospacer_surrogate_match_collapsed_zerocount": protospacer_surrogate_match_collapsed_zerocount,
                    "protospacer_surrogate_match_collapsed_gini": protospacer_surrogate_match_collapsed_gini,
                    "protospacer_surrogate_match_collapsed_coverage_mean": protospacer_surrogate_match_collapsed_coverage_mean,
                    "protospacer_surrogate_match_collapsed_coverage_std": protospacer_surrogate_match_collapsed_coverage_std,
                    f"protospacer_surrogate_match_collapsed_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_surrogate_match_collapsed_coverage_lower_quantile
                })
                if display_visualizations:
                    gini_index_bar_plot_labels.append("PROTOSPACER+SURROGATE")
                    collapsed_gini_index_bar_plot_values.append(protospacer_surrogate_match_collapsed_gini)

                    # Log warnings from count statistics
                    log_count_series_warnings(match_type_string="protospacer+surrogate-match",
                                      count_type_string="UMI-collapsed",
                                      total_count=protospacer_surrogate_match_collapsed_totalcount,
                                      protospacer_match_gt_given_match_type_total=protospacer_match_gt_protospacer_surrogate_match_collapsed_total,
                                      given_match_type_vs_protospacer_match_total_ratio=protospacer_surrogate_match_vs_protospacer_match_collapsed_total_ratio_threshold,
                                      given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                                      zerocount=protospacer_surrogate_match_collapsed_zerocount,
                                      gini=protospacer_surrogate_match_collapsed_gini,
                                      std=protospacer_surrogate_match_collapsed_coverage_std,
                                      gini_index_threshold=gini_index_threshold,
                                      coverage_lower_quantile=protospacer_surrogate_match_collapsed_coverage_lower_quantile,
                                      lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                                      lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                                      coverage_mean=protospacer_surrogate_match_collapsed_coverage_mean,
                                      mean_coverage_threshold=mean_coverage_threshold)

                # PROTOSPACER+SURROGATE vs. PROTOSPACER - NONCOLLAPSED
                protospacer_match_gt_protospacer_surrogate_match_noncollapsed_total = sum(count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries >= count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                protospacer_surrogate_match_vs_protospacer_match_noncollapsed_total_ratio_threshold = sum(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_noncollapsed_counterseries) / sum(count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                protospacer_surrogate_match_noncollapsed_zerocount = sum(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_noncollapsed_counterseries == 0)
                protospacer_surrogate_match_noncollapsed_totalcount = len(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                protospacer_surrogate_match_noncollapsed_gini = calculate_gini_index(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                protospacer_surrogate_match_noncollapsed_coverage_mean = np.mean(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                protospacer_surrogate_match_noncollapsed_coverage_std = np.std(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                protospacer_surrogate_match_noncollapsed_coverage_lower_quantile = count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_noncollapsed_counterseries.quantile(q=lower_quantile_coverage_percentage)

                protospacer_surrogate_match_umidups = count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_noncollapsed_counterseries / count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_collapsed_counterseries
                protospacer_surrogate_match_umidups_mean = protospacer_surrogate_match_umidups.mean()

                count_series_result_quality_control_stats_dict.update({
                    "protospacer_surrogate_match_noncollapsed_totalcount": protospacer_surrogate_match_noncollapsed_totalcount,
                    "protospacer_match_gt_protospacer_surrogate_match_noncollapsed_total": protospacer_match_gt_protospacer_surrogate_match_noncollapsed_total,
                    "protospacer_surrogate_match_vs_protospacer_match_noncollapsed_total_ratio_threshold": protospacer_surrogate_match_vs_protospacer_match_noncollapsed_total_ratio_threshold,
                    "protospacer_surrogate_match_noncollapsed_zerocount": protospacer_surrogate_match_noncollapsed_zerocount,
                    "protospacer_surrogate_match_noncollapsed_gini": protospacer_surrogate_match_noncollapsed_gini,
                    "protospacer_surrogate_match_noncollapsed_coverage_mean": protospacer_surrogate_match_noncollapsed_coverage_mean,
                    "protospacer_surrogate_match_noncollapsed_coverage_std": protospacer_surrogate_match_noncollapsed_coverage_std,
                    f"protospacer_surrogate_match_noncollapsed_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_surrogate_match_noncollapsed_coverage_lower_quantile,
                    "protospacer_surrogate_match_umidups_mean": protospacer_surrogate_match_umidups_mean
                })
                
                if display_visualizations:
                    noncollapsed_gini_index_bar_plot_values.append(protospacer_surrogate_match_noncollapsed_gini)

                    # Log warnings from count statistics
                    log_count_series_warnings(match_type_string="protospacer+surrogate-match",
                                      count_type_string="UMI-noncollapsed",
                                      total_count=protospacer_surrogate_match_noncollapsed_totalcount,
                                      protospacer_match_gt_given_match_type_total=protospacer_match_gt_protospacer_surrogate_match_noncollapsed_total,
                                      given_match_type_vs_protospacer_match_total_ratio=protospacer_surrogate_match_vs_protospacer_match_noncollapsed_total_ratio_threshold,
                                      given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                                      zerocount=protospacer_surrogate_match_noncollapsed_zerocount,
                                      gini=protospacer_surrogate_match_noncollapsed_gini,
                                      std=protospacer_surrogate_match_noncollapsed_coverage_std,
                                      gini_index_threshold=gini_index_threshold,
                                      coverage_lower_quantile=protospacer_surrogate_match_noncollapsed_coverage_lower_quantile,
                                      lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                                      lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                                      coverage_mean=protospacer_surrogate_match_noncollapsed_coverage_mean,
                                      mean_coverage_threshold=mean_coverage_threshold,
                                      umidups_mean=protospacer_surrogate_match_umidups_mean,
                                      umidups_lower_threshold=umidups_lower_threshold,
                                      umidups_upper_threshold=umidups_upper_threshold
                                      )
                

                if contains_barcode:
                        # PROTOSPACER+SURROGATE+BARCODE vs. PROTOSPACER - COLLAPSED
                        protospacer_match_gt_protospacer_surrogate_barcode_match_collapsed_total = sum(count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries >= count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries)
                        protospacer_surrogate_barcode_match_vs_protospacer_match_collapsed_total_ratio_threshold = sum(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries) / sum(count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries)
                        protospacer_surrogate_barcode_match_collapsed_zerocount = sum(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries == 0)
                        protospacer_surrogate_barcode_match_collapsed_totalcount = len(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries)
                        protospacer_surrogate_barcode_match_collapsed_gini = calculate_gini_index(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries)
                        protospacer_surrogate_barcode_match_collapsed_coverage_mean = np.mean(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries)
                        protospacer_surrogate_barcode_match_collapsed_coverage_std = np.std(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries)
                        protospacer_surrogate_barcode_match_collapsed_coverage_lower_quantile = count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries.quantile(q=lower_quantile_coverage_percentage)

                        count_series_result_quality_control_stats_dict.update({
                            "protospacer_surrogate_barcode_match_collapsed_totalcount": protospacer_surrogate_barcode_match_collapsed_totalcount,
                            "protospacer_match_gt_protospacer_surrogate_barcode_match_collapsed_total": protospacer_match_gt_protospacer_surrogate_barcode_match_collapsed_total,
                            "protospacer_surrogate_barcode_match_vs_protospacer_match_collapsed_total_ratio_threshold": protospacer_surrogate_barcode_match_vs_protospacer_match_collapsed_total_ratio_threshold,
                            "protospacer_surrogate_barcode_match_collapsed_zerocount": protospacer_surrogate_barcode_match_collapsed_zerocount,
                            "protospacer_surrogate_barcode_match_collapsed_gini": protospacer_surrogate_barcode_match_collapsed_gini,
                            "protospacer_surrogate_barcode_match_collapsed_coverage_mean": protospacer_surrogate_barcode_match_collapsed_coverage_mean,
                            "protospacer_surrogate_barcode_match_collapsed_coverage_std": protospacer_surrogate_barcode_match_collapsed_coverage_std,
                            f"protospacer_surrogate_barcode_match_collapsed_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_surrogate_barcode_match_collapsed_coverage_lower_quantile
                        })
                        
                        if display_visualizations:
                            gini_index_bar_plot_labels.append("PROTOSPACER+SURROGATE+BARCODE")
                            collapsed_gini_index_bar_plot_values.append(protospacer_surrogate_barcode_match_collapsed_gini)

                            # Log warnings from count statistics
                            log_count_series_warnings(match_type_string="protospacer+surrogate+barcode-match",
                                            count_type_string="UMI-collapsed",
                                            total_count=protospacer_surrogate_barcode_match_collapsed_totalcount,
                                            protospacer_match_gt_given_match_type_total=protospacer_match_gt_protospacer_surrogate_barcode_match_collapsed_total,
                                            given_match_type_vs_protospacer_match_total_ratio=protospacer_surrogate_barcode_match_vs_protospacer_match_collapsed_total_ratio_threshold,
                                            given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                                            zerocount=protospacer_surrogate_barcode_match_collapsed_zerocount,
                                            gini=protospacer_surrogate_barcode_match_collapsed_gini,
                                            std=protospacer_surrogate_barcode_match_collapsed_coverage_std,
                                            gini_index_threshold=gini_index_threshold,
                                            coverage_lower_quantile=protospacer_surrogate_barcode_match_collapsed_coverage_lower_quantile,
                                            lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                                            lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                                            coverage_mean=protospacer_surrogate_barcode_match_collapsed_coverage_mean,
                                            mean_coverage_threshold=mean_coverage_threshold)

                        # PROTOSPACER+SURROGATE+BARCODE vs. PROTOSPACER - NONCOLLAPSED
                        protospacer_match_gt_protospacer_surrogate_barcode_match_noncollapsed_total = sum(count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries >= count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                        protospacer_surrogate_barcode_match_vs_protospacer_match_noncollapsed_total_ratio_threshold = sum(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries) / sum(count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                        protospacer_surrogate_barcode_match_noncollapsed_zerocount = sum(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries == 0)
                        protospacer_surrogate_barcode_match_noncollapsed_totalcount = len(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                        protospacer_surrogate_barcode_match_noncollapsed_gini = calculate_gini_index(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                        protospacer_surrogate_barcode_match_noncollapsed_coverage_mean = np.mean(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                        protospacer_surrogate_barcode_match_noncollapsed_coverage_std = np.std(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                        protospacer_surrogate_barcode_match_noncollapsed_coverage_lower_quantile = count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries.quantile(q=lower_quantile_coverage_percentage)
                        
                        protospacer_surrogate_barcode_match_umidups = count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries / count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries
                        protospacer_surrogate_barcode_match_umidups_mean = protospacer_surrogate_barcode_match_umidups.mean()

                        count_series_result_quality_control_stats_dict.update({
                            "protospacer_surrogate_barcode_match_noncollapsed_totalcount": protospacer_surrogate_barcode_match_noncollapsed_totalcount,
                            "protospacer_match_gt_protospacer_surrogate_barcode_match_noncollapsed_total": protospacer_match_gt_protospacer_surrogate_barcode_match_noncollapsed_total,
                            "protospacer_surrogate_barcode_match_noncollapsed_zerocount": protospacer_surrogate_barcode_match_noncollapsed_zerocount,
                            "protospacer_surrogate_barcode_match_noncollapsed_gini": protospacer_surrogate_barcode_match_noncollapsed_gini,
                            "protospacer_surrogate_barcode_match_noncollapsed_coverage_mean": protospacer_surrogate_barcode_match_noncollapsed_coverage_mean,
                            "protospacer_surrogate_barcode_match_noncollapsed_coverage_std": protospacer_surrogate_barcode_match_noncollapsed_coverage_std,
                            f"protospacer_surrogate_barcode_match_noncollapsed_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_surrogate_barcode_match_noncollapsed_coverage_lower_quantile,
                            "protospacer_surrogate_barcode_match_umidups": protospacer_surrogate_barcode_match_umidups
                        })

                        if display_visualizations:
                            noncollapsed_gini_index_bar_plot_values.append(protospacer_surrogate_barcode_match_noncollapsed_gini)

                            # Log warnings from count statistics
                            log_count_series_warnings(match_type_string="protospacer+surrogate+barcode-match",
                                            count_type_string="UMI-noncollapsed",
                                            total_count=protospacer_surrogate_barcode_match_noncollapsed_totalcount,
                                            protospacer_match_gt_given_match_type_total=protospacer_match_gt_protospacer_surrogate_barcode_match_noncollapsed_total,
                                            given_match_type_vs_protospacer_match_total_ratio=protospacer_surrogate_barcode_match_vs_protospacer_match_noncollapsed_total_ratio_threshold,
                                            given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                                            zerocount=protospacer_surrogate_barcode_match_noncollapsed_zerocount,
                                            gini=protospacer_surrogate_barcode_match_noncollapsed_gini,
                                            std=protospacer_surrogate_barcode_match_noncollapsed_coverage_std,
                                            gini_index_threshold=protospacer_surrogate_barcode_match_noncollapsed_gini,
                                            coverage_lower_quantile=protospacer_surrogate_barcode_match_collapsed_coverage_lower_quantile,
                                            lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                                            lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                                            coverage_mean=protospacer_surrogate_barcode_match_noncollapsed_coverage_mean,
                                            mean_coverage_threshold=mean_coverage_threshold,
                                            umidups_mean=protospacer_surrogate_barcode_match_umidups,
                                            umidups_lower_threshold=umidups_lower_threshold,
                                            umidups_upper_threshold=umidups_upper_threshold)

        else:
            if contains_barcode:
                    # PROTOSPACER+BARCODE vs. PROTOSPACER - COLLAPSED
                    protospacer_match_gt_protospacer_barcode_match_collapsed_total = sum(count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries >= count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries)
                    protospacer_barcode_match_vs_protospacer_match_collapsed_total_ratio_threshold = sum(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries) / sum(count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries)
                    protospacer_barcode_match_collapsed_zerocount = sum(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries == 0)
                    protospacer_barcode_match_collapsed_totalcount = len(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries)
                    protospacer_barcode_match_collapsed_gini = calculate_gini_index(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries)
                    protospacer_barcode_match_collapsed_coverage_mean = np.mean(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries)
                    protospacer_barcode_match_collapsed_coverage_std = np.std(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries)
                    protospacer_barcode_match_collapsed_coverage_lower_quantile = count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries.quantile(q=lower_quantile_coverage_percentage)

                    count_series_result_quality_control_stats_dict.update({
                        "protospacer_barcode_match_collapsed_totalcount": protospacer_barcode_match_collapsed_totalcount,
                        "protospacer_match_gt_protospacer_barcode_match_collapsed_total": protospacer_match_gt_protospacer_barcode_match_collapsed_total,
                        "protospacer_barcode_match_vs_protospacer_match_collapsed_total_ratio_threshold": protospacer_barcode_match_vs_protospacer_match_collapsed_total_ratio_threshold,
                        "protospacer_barcode_match_collapsed_zerocount": protospacer_barcode_match_collapsed_zerocount,
                        "protospacer_barcode_match_collapsed_gini": protospacer_barcode_match_collapsed_gini,
                        "protospacer_barcode_match_collapsed_coverage_mean": protospacer_barcode_match_collapsed_coverage_mean,
                        "protospacer_barcode_match_collapsed_coverage_std": protospacer_barcode_match_collapsed_coverage_std,
                        f"protospacer_barcode_match_collapsed_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_barcode_match_collapsed_coverage_lower_quantile
                    })
                    
                    if display_visualizations:
                        gini_index_bar_plot_labels.append("PROTOSPACER+BARCODE")
                        collapsed_gini_index_bar_plot_values.append(protospacer_barcode_match_collapsed_gini)

                        # Log warnings from count statistics
                        log_count_series_warnings(match_type_string="protospacer+barcode-match",
                                        count_type_string="UMI-collapsed",
                                        total_count=protospacer_barcode_match_collapsed_totalcount,
                                        protospacer_match_gt_given_match_type_total=protospacer_match_gt_protospacer_barcode_match_collapsed_total,
                                        given_match_type_vs_protospacer_match_total_ratio=protospacer_barcode_match_vs_protospacer_match_collapsed_total_ratio_threshold,
                                        given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                                        zerocount=protospacer_barcode_match_collapsed_zerocount,
                                        gini=protospacer_barcode_match_collapsed_gini,
                                        std=protospacer_barcode_match_collapsed_coverage_std,
                                        gini_index_threshold=protospacer_barcode_match_collapsed_gini,
                                        coverage_lower_quantile=protospacer_barcode_match_collapsed_coverage_lower_quantile,
                                        lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                                        lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                                        coverage_mean=protospacer_barcode_match_collapsed_coverage_mean,
                                        mean_coverage_threshold=mean_coverage_threshold)



                    # PROTOSPACER+BARCODE vs. PROTOSPACER - NONCOLLAPSED
                    protospacer_match_gt_protospacer_barcode_match_noncollapsed_total = sum(count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries >= count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                    protospacer_barcode_match_vs_protospacer_match_noncollapsed_total_ratio_threshold = sum(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries) / sum(count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                    protospacer_barcode_match_noncollapsed_zerocount = sum(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries == 0)
                    protospacer_barcode_match_noncollapsed_totalcount = len(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                    protospacer_barcode_match_noncollapsed_gini = calculate_gini_index(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                    protospacer_barcode_match_noncollapsed_coverage_mean = np.mean(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                    protospacer_barcode_match_noncollapsed_coverage_std = np.std(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries)
                    protospacer_barcode_match_noncollapsed_coverage_lower_quantile = count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries.quantile(q=lower_quantile_coverage_percentage)

                    protospacer_barcode_match_umidups = count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries / count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries
                    protospacer_barcode_match_umidups_mean = protospacer_barcode_match_umidups.mean()

                    count_series_result_quality_control_stats_dict.update({
                        "protospacer_barcode_match_noncollapsed_totalcount": protospacer_barcode_match_noncollapsed_totalcount,
                        "protospacer_match_gt_protospacer_barcode_match_noncollapsed_total": protospacer_match_gt_protospacer_barcode_match_noncollapsed_total,
                        "protospacer_barcode_match_vs_protospacer_match_noncollapsed_total_ratio_threshold": protospacer_barcode_match_vs_protospacer_match_noncollapsed_total_ratio_threshold,
                        "protospacer_barcode_match_noncollapsed_zerocount": protospacer_barcode_match_noncollapsed_zerocount,
                        "protospacer_barcode_match_noncollapsed_gini": protospacer_barcode_match_noncollapsed_gini,
                        "protospacer_barcode_match_noncollapsed_coverage_mean": protospacer_barcode_match_noncollapsed_coverage_mean,
                        "protospacer_barcode_match_noncollapsed_coverage_std": protospacer_barcode_match_noncollapsed_coverage_std,
                        f"protospacer_barcode_match_noncollapsed_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_barcode_match_noncollapsed_coverage_lower_quantile,
                        "protospacer_barcode_match_umidups": protospacer_barcode_match_umidups
                    })

                    if display_visualizations:
                        noncollapsed_gini_index_bar_plot_values.append(protospacer_barcode_match_noncollapsed_gini)

                        # Log warnings from count statistics
                        log_count_series_warnings(match_type_string="protospacer+barcode-match",
                                        count_type_string="UMI-collapsed",
                                        total_count=protospacer_barcode_match_noncollapsed_totalcount,
                                        protospacer_match_gt_given_match_type_total=protospacer_match_gt_protospacer_barcode_match_noncollapsed_total,
                                        given_match_type_vs_protospacer_match_total_ratio=protospacer_barcode_match_vs_protospacer_match_noncollapsed_total_ratio_threshold,
                                        given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                                        zerocount=protospacer_barcode_match_noncollapsed_zerocount,
                                        gini=protospacer_barcode_match_noncollapsed_gini,
                                        std=protospacer_barcode_match_noncollapsed_coverage_std,
                                        gini_index_threshold=gini_index_threshold,
                                        coverage_lower_quantile=protospacer_barcode_match_noncollapsed_coverage_lower_quantile,
                                        lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                                        lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                                        coverage_mean=protospacer_barcode_match_noncollapsed_coverage_mean,
                                        mean_coverage_threshold=mean_coverage_threshold,
                                        umidups_mean=protospacer_barcode_match_umidups,
                                        umidups_lower_threshold=umidups_lower_threshold,
                                        umidups_upper_threshold=umidups_upper_threshold
                                        )

        # PLOT GINI INDEX:
        if display_visualizations:
            gini_index_bar_plot_input_df = pd.DataFrame({
                "UMI Collapsed": collapsed_gini_index_bar_plot_values,
                "UMI Non-collapsed": noncollapsed_gini_index_bar_plot_values}, index=gini_index_bar_plot_labels
                )
            gini_index_bar_plot_input_df.plot.bar(rot=0)
            plt.title("Gini-Index")
            plt.show()
    else:
        gini_index_bar_plot_labels: List[str] = []
        collapsed_gini_index_bar_plot_values: List[float] = []

        # PROTOSPACER+SURROGATE vs. PROTOSPACER
        protospacer_match_noumi_totalcount = len(count_series_result.protospacer_match.ambiguous_accepted_counterseries)
        protospacer_match_noumi_zerocount = sum(count_series_result.protospacer_match.ambiguous_accepted_counterseries == 0)
        protospacer_match_noumi_gini = calculate_gini_index(count_series_result.protospacer_match.ambiguous_accepted_counterseries)
        protospacer_match_noumi_coverage_mean = np.mean(count_series_result.protospacer_match.ambiguous_accepted_counterseries)
        protospacer_match_noumi_coverage_std = np.std(count_series_result.protospacer_match.ambiguous_accepted_counterseries)
        protospacer_match_noumi_coverage_lower_quantile = count_series_result.protospacer_match.ambiguous_accepted_counterseries.quantile(q=lower_quantile_coverage_percentage)

        count_series_result_quality_control_stats_dict.update({
            "protospacer_match_noumi_totalcount": protospacer_match_noumi_totalcount,
            "protospacer_match_noumi_zerocount": protospacer_match_noumi_zerocount,
            "protospacer_match_noumi_gini": protospacer_match_noumi_gini,
            "protospacer_match_noumi_coverage_mean": protospacer_match_noumi_coverage_mean,
            "protospacer_match_noumi_coverage_std": protospacer_match_noumi_coverage_std,
            f"protospacer_match_noumi_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_match_noumi_coverage_lower_quantile
        })
        if display_visualizations:
            # Log warnings from count statistics
            log_count_series_warnings(match_type_string="protospacer-match",
                            count_type_string="No-UMI",
                            total_count=None,
                            protospacer_match_gt_given_match_type_total=None,
                            given_match_type_vs_protospacer_match_total_ratio=None,
                            given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                            zerocount=protospacer_match_noumi_zerocount,
                            gini=protospacer_match_noumi_gini,
                            std=protospacer_match_noumi_coverage_std,
                            gini_index_threshold=gini_index_threshold,
                            coverage_lower_quantile=protospacer_match_noumi_coverage_lower_quantile,
                            lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                            lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                            coverage_mean=protospacer_match_noumi_coverage_mean,
                            mean_coverage_threshold=mean_coverage_threshold)
        if contains_surrogate:
            # PROTOSPACER+SURROGATE vs. PROTOSPACER
            protospacer_match_gt_protospacer_surrogate_match_noumi_total = sum(count_series_result.protospacer_match.ambiguous_accepted_counterseries >= count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_counterseries)
            protospacer_surrogate_match_vs_protospacer_match_noumi_total_ratio_threshold = sum(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_counterseries) / sum(count_series_result.protospacer_match.ambiguous_accepted_counterseries)
            protospacer_surrogate_match_noumi_totalcount = len(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_counterseries)
            protospacer_surrogate_match_noumi_zerocount = sum(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_counterseries == 0)
            protospacer_surrogate_match_noumi_gini = calculate_gini_index(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_counterseries)
            protospacer_surrogate_match_noumi_coverage_mean = np.mean(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_counterseries)
            protospacer_surrogate_match_noumi_coverage_std = np.std(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_counterseries)
            protospacer_surrogate_match_noumi_coverage_lower_quantile = count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_counterseries.quantile(q=lower_quantile_coverage_percentage)


            count_series_result_quality_control_stats_dict.update({
                "protospacer_surrogate_match_noumi_totalcount": protospacer_surrogate_match_noumi_totalcount,
                "protospacer_match_gt_protospacer_surrogate_match_noumi_total": protospacer_match_gt_protospacer_surrogate_match_noumi_total,
                "protospacer_surrogate_match_vs_protospacer_match_noumi_total_ratio_threshold": protospacer_surrogate_match_vs_protospacer_match_noumi_total_ratio_threshold,
                "protospacer_surrogate_match_noumi_zerocount": protospacer_surrogate_match_noumi_zerocount,
                "protospacer_surrogate_match_noumi_gini": protospacer_surrogate_match_noumi_gini,
                "protospacer_surrogate_match_noumi_coverage_mean": protospacer_surrogate_match_noumi_coverage_mean,
                "protospacer_surrogate_match_noumi_coverage_std": protospacer_surrogate_match_noumi_coverage_std,
                f"protospacer_surrogate_match_noumi_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_surrogate_match_noumi_coverage_lower_quantile
            })
            if display_visualizations:
                # Log warnings from count statistics
                log_count_series_warnings(match_type_string="protospacer+surrogate-match",
                                count_type_string="No-UMI",
                                total_count=protospacer_surrogate_match_noumi_totalcount,
                                protospacer_match_gt_given_match_type_total=protospacer_match_gt_protospacer_surrogate_match_noumi_total,
                                given_match_type_vs_protospacer_match_total_ratio=protospacer_surrogate_match_vs_protospacer_match_noumi_total_ratio_threshold,
                                given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                                zerocount=protospacer_surrogate_match_noumi_zerocount,
                                gini=protospacer_surrogate_match_noumi_gini,
                                std=protospacer_surrogate_match_noumi_coverage_std,
                                gini_index_threshold=gini_index_threshold,
                                coverage_lower_quantile=protospacer_surrogate_match_noumi_coverage_lower_quantile,
                                lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                                lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                                coverage_mean=protospacer_surrogate_match_noumi_coverage_mean,
                                mean_coverage_threshold=mean_coverage_threshold)

            if contains_barcode:
                # PROTOSPACER+SURROGATE+BARCODE vs. PROTOSPACER
                protospacer_match_gt_protospacer_surrogate_barcode_match_noumi_total = sum(count_series_result.protospacer_match.ambiguous_accepted_counterseries >= count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_counterseries)
                protospacer_surrogate_barcode_match_vs_protospacer_match_noumi_total_ratio_threshold = sum(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_counterseries) / sum(count_series_result.protospacer_match.ambiguous_accepted_counterseries)
                protospacer_surrogate_barcode_match_noumi_totalcount = len(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_counterseries)
                protospacer_surrogate_barcode_match_noumi_zerocount = sum(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_counterseries == 0)
                protospacer_surrogate_barcode_match_noumi_gini = calculate_gini_index(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_counterseries)
                protospacer_surrogate_barcode_match_noumi_coverage_mean = np.mean(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_counterseries)
                protospacer_surrogate_barcode_match_noumi_coverage_std = np.std(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_counterseries)
                protospacer_surrogate_barcode_match_noumi_coverage_lower_quantile = count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_counterseries.quantile(q=lower_quantile_coverage_percentage)

                count_series_result_quality_control_stats_dict.update({
                    "protospacer_surrogate_barcode_match_noumi_totalcount": protospacer_surrogate_barcode_match_noumi_totalcount,
                    "protospacer_match_gt_protospacer_surrogate_barcode_match_noumi_total": protospacer_match_gt_protospacer_surrogate_barcode_match_noumi_total,
                    "protospacer_surrogate_barcode_match_vs_protospacer_match_noumi_total_ratio_threshold": protospacer_surrogate_barcode_match_vs_protospacer_match_noumi_total_ratio_threshold,
                    "protospacer_surrogate_barcode_match_noumi_zerocount": protospacer_surrogate_barcode_match_noumi_zerocount,
                    "protospacer_surrogate_barcode_match_noumi_gini": protospacer_surrogate_barcode_match_noumi_gini,
                    "protospacer_surrogate_barcode_match_noumi_coverage_mean": protospacer_surrogate_barcode_match_noumi_coverage_mean,
                    "protospacer_surrogate_barcode_match_noumi_coverage_std": protospacer_surrogate_barcode_match_noumi_coverage_std,
                    f"protospacer_surrogate_barcode_match_noumi_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_surrogate_barcode_match_noumi_coverage_lower_quantile
                })

                if display_visualizations:
                    # Log warnings from count statistics
                    log_count_series_warnings(match_type_string="protospacer+surrogate+barcode-match",
                                    count_type_string="No-UMI",
                                    total_count=protospacer_surrogate_barcode_match_noumi_totalcount,
                                    protospacer_match_gt_given_match_type_total=protospacer_match_gt_protospacer_surrogate_barcode_match_noumi_total,
                                    given_match_type_vs_protospacer_match_total_ratio=protospacer_surrogate_barcode_match_vs_protospacer_match_noumi_total_ratio_threshold,
                                    given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                                    zerocount=protospacer_surrogate_match_noumi_zerocount,
                                    gini=protospacer_surrogate_match_noumi_gini,
                                    std=protospacer_surrogate_match_noumi_coverage_std,
                                    gini_index_threshold=gini_index_threshold,
                                    coverage_lower_quantile=protospacer_surrogate_match_noumi_coverage_lower_quantile,
                                    lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                                    lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                                    coverage_mean=protospacer_surrogate_match_noumi_coverage_mean,
                                    mean_coverage_threshold=mean_coverage_threshold)

        else:
            if contains_barcode:
                # PROTOSPACER+BARCODE vs. PROTOSPACER
                protospacer_match_gt_protospacer_barcode_match_noumi_total = sum(count_series_result.protospacer_match.ambiguous_accepted_counterseries >= count_series_result.protospacer_match_barcode_match.ambiguous_accepted_counterseries)
                protospacer_barcode_match_vs_protospacer_match_noumi_total_ratio_threshold = sum(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_counterseries) / sum(count_series_result.protospacer_match.ambiguous_accepted_counterseries)
                protospacer_barcode_match_noumi_totalcount = len(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_counterseries)
                protospacer_barcode_match_noumi_zerocount = sum(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_counterseries == 0)
                protospacer_barcode_match_noumi_gini = calculate_gini_index(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_counterseries)
                protospacer_barcode_match_noumi_coverage_mean = np.mean(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_counterseries)
                protospacer_barcode_match_noumi_coverage_std = np.std(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_counterseries)
                protospacer_barcode_match_noumi_coverage_lower_quantile = count_series_result.protospacer_match_barcode_match.ambiguous_accepted_counterseries.quantile(q=lower_quantile_coverage_percentage)

                count_series_result_quality_control_stats_dict.update({
                    "protospacer_barcode_match_noumi_totalcount": protospacer_barcode_match_noumi_totalcount,
                    "protospacer_match_gt_protospacer_barcode_match_noumi_total": protospacer_match_gt_protospacer_barcode_match_noumi_total,
                    "protospacer_barcode_match_vs_protospacer_match_noumi_total_ratio_threshold": protospacer_barcode_match_vs_protospacer_match_noumi_total_ratio_threshold,
                    "protospacer_barcode_match_noumi_zerocount": protospacer_barcode_match_noumi_zerocount,
                    "protospacer_barcode_match_noumi_gini": protospacer_barcode_match_noumi_gini,
                    "protospacer_barcode_match_noumi_coverage_mean": protospacer_barcode_match_noumi_coverage_mean,
                    "protospacer_barcode_match_noumi_coverage_std": protospacer_barcode_match_noumi_coverage_std,
                    f"protospacer_barcode_match_noumi_coverage_lower_quantile_{lower_quantile_coverage_percentage}": protospacer_barcode_match_noumi_coverage_lower_quantile
                })

                if display_visualizations:
                    # Log warnings from count statistics
                    log_count_series_warnings(match_type_string="protospacer+barcode-match",
                                    count_type_string="No-UMI",
                                    total_count=protospacer_barcode_match_noumi_totalcount,
                                    protospacer_match_gt_given_match_type_total=protospacer_match_gt_protospacer_barcode_match_noumi_total,
                                    given_match_type_vs_protospacer_match_total_ratio=protospacer_barcode_match_vs_protospacer_match_noumi_total_ratio_threshold,
                                    given_match_type_vs_protospacer_match_total_ratio_threshold=given_match_type_vs_protospacer_match_total_ratio_threshold,
                                    zerocount=protospacer_barcode_match_noumi_zerocount,
                                    gini=protospacer_barcode_match_noumi_gini,
                                    std=protospacer_barcode_match_noumi_coverage_std,
                                    gini_index_threshold=gini_index_threshold,
                                    coverage_lower_quantile=protospacer_barcode_match_noumi_coverage_lower_quantile,
                                    lower_quantile_coverage_percentage=lower_quantile_coverage_percentage,
                                    lower_quantile_coverage_threshold=lower_quantile_coverage_threshold,
                                    coverage_mean=protospacer_barcode_match_noumi_coverage_mean,
                                    mean_coverage_threshold=mean_coverage_threshold)
    return count_series_result_quality_control_stats_dict
    

def plot_match_type_comparisons(count_series_result: MatchSetWhitelistReporterCounterSeriesResults, contains_umi: bool, contains_surrogate: bool, contains_barcode: bool, collapsed:bool=False, alpha: float=0.3):
    if contains_umi:
        if contains_surrogate:
            if collapsed:
                # PROTOSPACER+SURROGATE vs. PROTOSPACER - COLLAPSED
                plt.scatter(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_collapsed_counterseries,
                        count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries,
                    alpha=alpha)
                plt.xlabel("Protospacer+Surrogate Match")
                plt.ylabel("Protospacer Match")
                plt.title("Comparison of COLLAPSED Protospacer+Surrogate Match vs. Protospacer-only match")
                plt.show()
            else:
                # PROTOSPACER+SURROGATE vs. PROTOSPACER - NONCOLLAPSED
                plt.scatter(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_umi_noncollapsed_counterseries,
                        count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries,
                    alpha=alpha)
                plt.xlabel("Protospacer+Surrogate Match")
                plt.ylabel("Protospacer Match")
                plt.title("Comparison of NON-COLLAPSED Protospacer+Surrogate Match vs. Protospacer-only match")
                plt.show()

            if contains_barcode:
                if collapsed:
                    # PROTOSPACER+SURROGATE+BARCODE vs. PROTOSPACER - COLLAPSED
                    plt.scatter(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries,
                        count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries,
                    alpha=alpha)
                    plt.xlabel("Protospacer+Surrogate+Barcode Match")
                    plt.ylabel("Protospacer Match")
                    plt.title("Comparison of COLLAPSED Protospacer+Surrogate+Barcode Match vs. Protospacer-only match")
                    plt.show()
                else:
                    # PROTOSPACER+SURROGATE+BARCODE vs. PROTOSPACER - NONCOLLAPSED
                    plt.scatter(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries,
                        count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries,
                    alpha=alpha)
                    plt.xlabel("Protospacer+Surrogate+Barcode Match")
                    plt.ylabel("Protospacer Match")
                    plt.title("Comparison of NON-COLLAPSED Protospacer+Surrogate+Barcode Match vs. Protospacer-only match")
                    plt.show()
        else:
            if contains_barcode:
                if collapsed:
                    # PROTOSPACER+BARCODE vs. PROTOSPACER - COLLAPSED
                    plt.scatter(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_collapsed_counterseries,
                        count_series_result.protospacer_match.ambiguous_accepted_umi_collapsed_counterseries,
                    alpha=alpha)
                    plt.xlabel("Protospacer+Barcode Match")
                    plt.ylabel("Protospacer Match")
                    plt.title("Comparison of COLLAPSED Protospacer+Barcode Match vs. Protospacer-only match")
                    plt.show()
                else:
                    # PROTOSPACER+BARCODE vs. PROTOSPACER - NONCOLLAPSED
                    plt.scatter(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_umi_noncollapsed_counterseries,
                        count_series_result.protospacer_match.ambiguous_accepted_umi_noncollapsed_counterseries,
                    alpha=alpha)
                    plt.xlabel("Protospacer+Barcode Match")
                    plt.ylabel("Protospacer Match")
                    plt.title("Comparison of NON-COLLAPSED Protospacer+Barcode Match vs. Protospacer-only match")
                    plt.show()
    else:

        if contains_surrogate:
            # PROTOSPACER+SURROGATE vs. PROTOSPACER
            plt.scatter(count_series_result.protospacer_match_surrogate_match.ambiguous_accepted_counterseries,
                    count_series_result.protospacer_match.ambiguous_accepted_counterseries,
                alpha=alpha)
            plt.xlabel("Protospacer+Surrogate Match")
            plt.ylabel("Protospacer Match")
            plt.title("Comparison of Protospacer+Surrogate Match vs. Protospacer-only match")
            plt.show()
            
            if contains_barcode:

                # PROTOSPACER+SURROGATE+BARCODE vs. PROTOSPACER
                plt.scatter(count_series_result.protospacer_match_surrogate_match_barcode_match.ambiguous_accepted_counterseries,
                    count_series_result.protospacer_match.ambiguous_accepted_counterseries,
                alpha=alpha)
                plt.xlabel("Protospacer+Surrogate+Barcode Match")
                plt.ylabel("Protospacer Match")
                plt.title("Comparison of Protospacer+Surrogate+Barcode Match vs. Protospacer-only match")
                plt.show()
        else:
            if contains_barcode:

                # PROTOSPACER+BARCODE vs. PROTOSPACER
                plt.scatter(count_series_result.protospacer_match_barcode_match.ambiguous_accepted_counterseries,
                    count_series_result.protospacer_match.ambiguous_accepted_counterseries,
                alpha=alpha)
                plt.xlabel("Protospacer+Barcode Match")
                plt.ylabel("Protospacer Match")
                plt.title("Comparison of Protospacer+Barcode Match vs. Protospacer-only match")
                plt.show()



from typeguard import typechecked
from typing import Union, List, Mapping, Tuple, Optional, Any, DefaultDict, Dict
from typing import Counter as CounterType
from ..models.mapping_models import GeneralMappingInferenceDict, GeneralMatchCountDict, GeneralMismatchCountDict, WhitelistReporterObservedSequenceMapping

from collections import Counter
from collections import defaultdict
from ..models.mapping_models import MatchSetWhitelistReporterCounterSeriesResults, CompleteInferenceMatchResult, MatchSetSingleInferenceMatchResult, SurrogateProtospacerMismatchSetWhitelistReporterCounterSeriesResults, SurrogateProtospacerMismatchSingleInferenceMatchResult, AllMatchSetWhitelistReporterCounterSeriesResults, InferenceResult
from .crispr_editing_processing import check_match_result_non_error, get_non_error_dict
import pandas as pd

@typechecked
def helper_get_observed_values_given_whitelist_value(whitelist_sequence_list: List[Tuple[str, Optional[str], Optional[str]]], observed_guide_reporter_umi_counts_inferred: GeneralMappingInferenceDict, attribute_name:str, contains_umi: bool, ambiguous_accepted:bool = True) -> WhitelistReporterObservedSequenceMapping:
    """
    Provides the set of observed sequences for a given whitelist sequence

    Args:
        whitelist_sequence_list (List[Tuple[str, Optional[str], Optional[str]]]): List of whitelist sequence to retrieve observed sequences
        observed_guide_reporter_umi_counts_inferred (GeneralMappingInferenceDict): Datastructure contain the observed sequence to mapped whitelist sequence
        attribute_name (str): string specifying white mapping match type to consider.
        contains_umi (bool): specifying whether UMI is used
        ambiguous_accepted (bool): specifying whether to consider ambiguous mapping

    Returns:
        WhitelistReporterObservedSequenceMapping: dictionary mapping between the whitelist sequence and the list of observed sequence. Each observed sequence has the corresponding count Union[int, Dict [ str, int ]] (either int if no UMI, or Dict[str, int] for UMI-collapsed and non-collpased count)
    """
    whitelist_sequence_mapping_list: WhitelistReporterObservedSequenceMapping = defaultdict(list)


    # Iterate through the inference results (will attempt to do retrieval for all requested whitelist sequences at once for optimization)
    inferred_value_results: InferenceResult
    observed_sequence: Tuple[str, Optional[str], Optional[str]]
    for observed_sequence, inferred_value_results in get_non_error_dict(observed_guide_reporter_umi_counts_inferred, attribute_name).items():
        #
        #   Get the relevant attributes
        #
        observed_value_counts: Union[int, CounterType[Optional[str]]] = inferred_value_results.observed_value # Read count of the observed sequence
        inferred_value_result: CompleteInferenceMatchResult =  inferred_value_results.inferred_value  # Whitelist inference results
        match_set_single_inference_match_result : Optional[MatchSetSingleInferenceMatchResult] = getattr(inferred_value_result, attribute_name) # Get inference result for specific mapping strategy
        assert match_set_single_inference_match_result is not None, "match_set_single_inference_match_result should not be none since this is from the non error list. Developer error."
        
        matches: pd.DataFrame = match_set_single_inference_match_result.value.matches # Get the list of ambiguous
        
        if not matches.empty:
            
            # Skip observed sequence if there are multiple matches when not accepting ambiguous mapping 
            if (ambiguous_accepted is False) and (matches.shape[0] > 1):
                continue
            
            
            for whitelist_reporter_series in matches.iterrows(): 
                # UMI-BASED COUNTING
                dict_index = tuple(whitelist_reporter_series[1])
                if dict_index in whitelist_sequence_list: # If the match is in the requested whitelist sequences, proceed
                    # Get match info to add to result
                    all_match_sequences = [tuple(whitelist_reporter_series[1]) for whitelist_reporter_series in matches.iterrows()]
                    total_match_counts = len(all_match_sequences)

                    if contains_umi:
                        assert isinstance(observed_value_counts, Counter), f"For UMI, expecting observed value is a Counter, but type is {type(observed_value_counts)}"

                        # Calculate both UMI noncollapsed and collapsed count
                        observed_sequence_count = {
                            "umi_noncollapsed_count": sum(observed_value_counts.values()),
                            "umi_collapsed_count": len(observed_value_counts.values())
                        }

                        # Add observed sequence and count to mapping
                        whitelist_sequence_mapping_list[dict_index].append( (observed_sequence, observed_sequence_count, total_match_counts, all_match_sequences) )
                    
                    # STANDARD NON-UMI BASED COUNTING
                    else:
                        assert isinstance(observed_value_counts, int), f"For non UMI, expecting observed value is an int, but type is {type(observed_value_counts)}"
                        
                        observed_sequence_count = observed_value_counts
                        
                        # Add observed sequence and count to mapping
                        whitelist_sequence_mapping_list[dict_index].append( (observed_sequence, observed_sequence_count, total_match_counts, all_match_sequences) )

    return whitelist_sequence_mapping_list



#
#   GET THE WHITELIST COUNT PANDAS SERIES
#
# HELPER FUNCTION GETS COUNTS FOR THE THE MATCHES - defined in-function to reduce arguments being passed (NOTE: There is some duplicate code with mismatch counts function - keep in mind if making modifications)
@typechecked
def get_matchset_counterseries(observed_guide_reporter_umi_counts_inferred: GeneralMappingInferenceDict, whitelist_guide_reporter_df: pd.DataFrame, contains_umi: bool, attribute_name: str) -> MatchSetWhitelistReporterCounterSeriesResults: 
    #
    #   DEFINE THE DEFAULTDICTS FOR COUNTING
    #
    ambiguous_ignored_umi_noncollapsed_counterdict : GeneralMatchCountDict = defaultdict(int)
    ambiguous_ignored_umi_collapsed_counterdict : GeneralMatchCountDict  = defaultdict(int)
    ambiguous_ignored_counterdict : GeneralMatchCountDict  = defaultdict(int)

    ambiguous_accepted_umi_noncollapsed_counterdict : GeneralMatchCountDict  = defaultdict(int)
    ambiguous_accepted_umi_collapsed_counterdict : GeneralMatchCountDict  = defaultdict(int)
    ambiguous_accepted_counterdict : GeneralMatchCountDict  = defaultdict(int)

    ambiguous_spread_umi_noncollapsed_counterdict : GeneralMatchCountDict  = defaultdict(float)
    ambiguous_spread_umi_collapsed_counterdict : GeneralMatchCountDict  = defaultdict(float)
    ambiguous_spread_counterdict : GeneralMatchCountDict  = defaultdict(float)

    #
    # ITERATE THROUGH THE NON-ERROR INFERRED RESULTS AND FILL THE COUNTS
    #
    inferred_value_results: InferenceResult
    for inferred_value_results in get_non_error_dict(observed_guide_reporter_umi_counts_inferred, attribute_name).values():
        #
        #   Get the relevant attributes
        #
        observed_value_counts: Union[int, CounterType[Optional[str]]] = inferred_value_results.observed_value
        inferred_value_result: CompleteInferenceMatchResult =  inferred_value_results.inferred_value 
        match_set_single_inference_match_result : Optional[MatchSetSingleInferenceMatchResult] = getattr(inferred_value_result, attribute_name)
        assert match_set_single_inference_match_result is not None, "match_set_single_inference_match_result should not be none since this is from the non error list. Developer error."
        
        matches: pd.DataFrame = match_set_single_inference_match_result.value.matches
        if not matches.empty:
            # ITERATE THROUGH MATCHE(S) TO PERFORM COUNTS
            for whitelist_reporter_series in matches.iterrows(): 
                # UMI-BASED COUNTING
                dict_index = tuple(whitelist_reporter_series[1])
                if contains_umi:
                    assert isinstance(observed_value_counts, Counter), f"For UMI, expecting observed value is a Counter, but type is {type(observed_value_counts)}"
                    ambiguous_accepted_umi_noncollapsed_counterdict[dict_index] += sum(observed_value_counts.values())
                    ambiguous_accepted_umi_collapsed_counterdict[dict_index] += len(observed_value_counts.values())

                    ambiguous_spread_umi_noncollapsed_counterdict[dict_index] += sum(observed_value_counts.values()) / float(matches.shape[0])
                    ambiguous_spread_umi_collapsed_counterdict[dict_index] += len(observed_value_counts.values()) / float(matches.shape[0])
                    
                    # If there is no ambiguous matches, then add to ambiguous_ignored counter
                    if matches.shape[0] == 1:
                        ambiguous_ignored_umi_noncollapsed_counterdict[dict_index] += sum(observed_value_counts.values())
                        ambiguous_ignored_umi_collapsed_counterdict[dict_index] += len(observed_value_counts.values())
                
                # STANDARD NON-UMI BASED COUNTING
                else:
                    assert isinstance(observed_value_counts, int), f"For non UMI, expecting observed value is an int, but type is {type(observed_value_counts)}"
                    ambiguous_accepted_counterdict[dict_index] += observed_value_counts
                    ambiguous_spread_counterdict[dict_index] += observed_value_counts / float(matches.shape[0])
                    
                    # If there is no ambiguous matches, then add to ambiguous_ignored counter
                    if matches.shape[0] == 1:
                        ambiguous_ignored_counterdict[dict_index] += observed_value_counts
    
    # Helper function that converts defaultdict to series
    def create_counterseries(counterdict: GeneralMatchCountDict) -> pd.Series:
        counterseries: pd.Series = whitelist_guide_reporter_df.apply(lambda reporter: counterdict[tuple(reporter)], axis=1)
        counterseries.index = pd.MultiIndex.from_frame(whitelist_guide_reporter_df)
        return counterseries
    

    #
    #   CONVERT THE COUNT DICTS INTO PANDAS SERIES, since this is a more ideal structure.
    #
    match_set_whitelist_reporter_counter_series_results = MatchSetWhitelistReporterCounterSeriesResults()
    match_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_noncollapsed_counterseries = create_counterseries(ambiguous_ignored_umi_noncollapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_collapsed_counterseries = create_counterseries(ambiguous_ignored_umi_collapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_ignored_counterseries = create_counterseries(ambiguous_ignored_counterdict)

    match_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_noncollapsed_counterseries = create_counterseries(ambiguous_accepted_umi_noncollapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_collapsed_counterseries = create_counterseries(ambiguous_accepted_umi_collapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_accepted_counterseries = create_counterseries(ambiguous_accepted_counterdict)

    match_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_noncollapsed_counterseries = create_counterseries(ambiguous_spread_umi_noncollapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_collapsed_counterseries = create_counterseries(ambiguous_spread_umi_collapsed_counterdict)
    match_set_whitelist_reporter_counter_series_results.ambiguous_spread_counterseries = create_counterseries(ambiguous_spread_counterdict)

    return match_set_whitelist_reporter_counter_series_results


#
# HELPER FUNCTION GETS COUNTS FOR THE THE MISMATCHES - defined in-function to reduce arguments being passed (NOTE: There is some duplicate code with match counts function - keep in mind if making modifications)
#
@typechecked
def get_mismatchset_counterseries(observed_guide_reporter_umi_counts_inferred: GeneralMappingInferenceDict, whitelist_guide_reporter_df: pd.DataFrame, contains_umi: bool, attribute_name: str) -> SurrogateProtospacerMismatchSetWhitelistReporterCounterSeriesResults:
    #
    #   DEFINE THE DEFAULTDICTS FOR COUNTING
    #
    # MATCH counters
    ambiguous_ignored_umi_noncollapsed_match_counterdict : GeneralMatchCountDict  = defaultdict(int)
    ambiguous_ignored_umi_collapsed_match_counterdict : GeneralMatchCountDict  = defaultdict(int)
    ambiguous_ignored_match_counterdict : GeneralMatchCountDict  = defaultdict(int)

    ambiguous_accepted_umi_noncollapsed_match_counterdict : GeneralMatchCountDict  = defaultdict(int)
    ambiguous_accepted_umi_collapsed_match_counterdict : GeneralMatchCountDict  = defaultdict(int)
    ambiguous_accepted_match_counterdict : GeneralMatchCountDict  = defaultdict(int)

    ambiguous_spread_umi_noncollapsed_match_counterdict : GeneralMatchCountDict  = defaultdict(float)
    ambiguous_spread_umi_collapsed_match_counterdict : GeneralMatchCountDict  = defaultdict(float)
    ambiguous_spread_match_counterdict : GeneralMatchCountDict  = defaultdict(float)

    # MISMATCH counters (keys are PAIRS of indices, representing the protospacer and surrogate match separately)
    ambiguous_ignored_umi_noncollapsed_mismatch_counterdict : GeneralMismatchCountDict  = defaultdict(int)
    ambiguous_ignored_umi_collapsed_mismatch_counterdict : GeneralMismatchCountDict  = defaultdict(int)
    ambiguous_ignored_mismatch_counterdict : GeneralMismatchCountDict  = defaultdict(int)

    ambiguous_accepted_umi_noncollapsed_mismatch_counterdict : GeneralMismatchCountDict  = defaultdict(int)
    ambiguous_accepted_umi_collapsed_mismatch_counterdict : GeneralMismatchCountDict  = defaultdict(int)
    ambiguous_accepted_mismatch_counterdict : GeneralMismatchCountDict  = defaultdict(int)

    ambiguous_spread_umi_noncollapsed_mismatch_counterdict : GeneralMismatchCountDict  = defaultdict(float)
    ambiguous_spread_umi_collapsed_mismatch_counterdict : GeneralMismatchCountDict = defaultdict(float)
    ambiguous_spread_mismatch_counterdict : GeneralMismatchCountDict = defaultdict(float)

    #
    # ITERATE THROUGH THE NON-ERROR INFERRED RESULTS (NOTE: If only one of the protospacer or surrogate matches but not the other, this is treated as an error, and will NOT be counted or considered (even in the single match series). For those counts, should just use the protospacer-only or surrogate-only match results)
    #
    inferred_value_results: InferenceResult
    for inferred_value_results in get_non_error_dict(observed_guide_reporter_umi_counts_inferred, attribute_name).values():
        #
        #   Get the relevant attributes
        #
        observed_value_counts: Union[int, CounterType[Optional[str]]] = inferred_value_results.observed_value
        inferred_value_result: CompleteInferenceMatchResult =  inferred_value_results.inferred_value
        surrogate_protospacer_mismatch_single_inference_match_result : Optional[SurrogateProtospacerMismatchSingleInferenceMatchResult] = getattr(inferred_value_result, attribute_name)
        assert surrogate_protospacer_mismatch_single_inference_match_result is not None, "surrogate_protospacer_mismatch_single_inference_match_result should not be none since this is from the non error list. Developer error."
        
        mismatched: bool = surrogate_protospacer_mismatch_single_inference_match_result.value.mismatched
        surrogate_matches: pd.DataFrame = surrogate_protospacer_mismatch_single_inference_match_result.value.surrogate_matches
        protospacer_matches: pd.DataFrame = surrogate_protospacer_mismatch_single_inference_match_result.value.protospacer_matches
        protospacer_surrogate_matches: pd.DataFrame = surrogate_protospacer_mismatch_single_inference_match_result.value.protospacer_surrogate_matches



        if mismatched: # If mismatched, Add count to protospacer/surrogate pair. NOTE: Both the protospacer/surrogate matches should be gauranteed
            assert not surrogate_matches.empty, "Developer error: to be called a mismatch, there must be both separate protospacer and surrogate. No surrogate match (possible no protospacer match)"
            assert not protospacer_matches.empty, "Developer error: to be called a mismatch, there must be both separate protospacer and surrogate. No surrogate match (possible no protospacer match)"
            assert protospacer_surrogate_matches.empty, "Developer error: to be called a mismatch, matches dataframe must be empty."

            # NOTE: For the mismatches, we want to tally for the specific protospacer_match/surrogate_match pair, so we iterate through both match dataframes to tally
            for protospacer_matched_whitelist_reporter_series in protospacer_matches.iterrows():
                for surrogate_matched_whitelist_reporter_series in surrogate_matches.iterrows():
                    
                    # NOTE: For the counter_dict key, the order of the pairwise tuple is PROTOSPACER_MATCH, SURROGATE_MATCH.
                    dict_index = (tuple(protospacer_matched_whitelist_reporter_series[1]), tuple(surrogate_matched_whitelist_reporter_series[1]))
                    
                    if contains_umi:
                        assert isinstance(observed_value_counts, Counter), f"For UMI, expecting observed value is a Counter, but type is {type(observed_value_counts)}"
                        
                        ambiguous_accepted_umi_noncollapsed_mismatch_counterdict[dict_index] += sum(observed_value_counts.values())
                        ambiguous_accepted_umi_collapsed_mismatch_counterdict[dict_index] += len(observed_value_counts.values())

                        ambiguous_spread_umi_noncollapsed_mismatch_counterdict[dict_index] += sum(observed_value_counts.values()) / float(protospacer_matches.shape[0]*surrogate_matches.shape[0]) # NOTE: The denominator for the "spreaded count" is intuitively the number of pairs from the protospacer/surrogate-matches, so just the product
                        ambiguous_spread_umi_collapsed_mismatch_counterdict[dict_index] += len(observed_value_counts.values()) / float(protospacer_matches.shape[0]*surrogate_matches.shape[0])
                        
                        # If there is no ambiguous matches, then add to ambiguous_ignored counter
                        if (protospacer_matches.shape[0] == 1) and (surrogate_matches.shape[0] == 1):
                            ambiguous_ignored_umi_noncollapsed_mismatch_counterdict[dict_index] += sum(observed_value_counts.values())
                            ambiguous_ignored_umi_collapsed_mismatch_counterdict[dict_index] += len(observed_value_counts.values())
                    else:
                        assert isinstance(observed_value_counts, int), f"For non UMI, expecting observed value is an int, but type is {type(observed_value_counts)}"

                        ambiguous_accepted_mismatch_counterdict[dict_index] += observed_value_counts
                        ambiguous_spread_mismatch_counterdict[dict_index] += observed_value_counts / float(protospacer_matches.shape[0]*surrogate_matches.shape[0])
                        
                        # If there is no ambiguous matches, then add to ambiguous_ignored counter
                        if (protospacer_matches.shape[0] == 1) and (surrogate_matches.shape[0] == 1):
                            ambiguous_ignored_mismatch_counterdict[dict_index] += observed_value_counts


        else: # If matched, add count normally.(NOTE: As mentioned above, "If only one of the protospacer or surrogate matches but not the other, this is treated as an error, and will NOT be counted or considered (even in the single match series). For those counts, should just use the protospacer-only or surrogate-only match results")
            assert not protospacer_surrogate_matches.empty, f"mismatched==false, but the match dataframe is empty, unexpected paradox. Developer error"
            
            matches = protospacer_surrogate_matches
            # NOTE/TODO: FROM HERE IS THE SAME COUNTING LOGIC AS THE MATCH COUNTING FUNCTION - can modularize.
            for whitelist_reporter_series in matches.iterrows():
                dict_index = tuple(whitelist_reporter_series[1])
                if contains_umi:
                    assert isinstance(observed_value_counts, Counter), f"For UMI, expecting observed value is a Counter, but type is {type(observed_value_counts)}"
                    
                    ambiguous_accepted_umi_noncollapsed_match_counterdict[dict_index] += sum(observed_value_counts.values())
                    ambiguous_accepted_umi_collapsed_match_counterdict[dict_index] += len(observed_value_counts.values())

                    ambiguous_spread_umi_noncollapsed_match_counterdict[dict_index] += sum(observed_value_counts.values()) / float(matches.shape[0])
                    ambiguous_spread_umi_collapsed_match_counterdict[dict_index] += len(observed_value_counts.values()) / float(matches.shape[0])
                    
                    # If there is no ambiguous matches, then add to ambiguous_ignored counter
                    if matches.shape[0] == 1:
                        ambiguous_ignored_umi_noncollapsed_match_counterdict[dict_index] += sum(observed_value_counts.values())
                        ambiguous_ignored_umi_collapsed_match_counterdict[dict_index] += len(observed_value_counts.values())
                else:
                    assert isinstance(observed_value_counts, int), f"For non UMI, expecting observed value is an int, but type is {type(observed_value_counts)}"
                    ambiguous_accepted_match_counterdict[dict_index] += observed_value_counts
                    ambiguous_spread_match_counterdict[dict_index] += observed_value_counts / float(matches.shape[0])
                    
                    # If there is no ambiguous matches, then add to ambiguous_ignored counter
                    if matches.shape[0] == 1:
                        ambiguous_ignored_match_counterdict[dict_index] += observed_value_counts
    
    def create_match_counterseries(counterdict: DefaultDict[Tuple[str, Optional[str], Optional[str]], Union[int, float]]) -> pd.Series:
        counterseries: pd.Series = whitelist_guide_reporter_df.apply(lambda reporter: counterdict[tuple(reporter)], axis=1)
        counterseries.index = pd.MultiIndex.from_frame(whitelist_guide_reporter_df)
        return counterseries
    
    def create_mismatch_counterseries(counterdict: DefaultDict[Tuple[Tuple[str, Optional[str], Optional[str]], Tuple[str, Optional[str], Optional[str]]], Union[int, float]]) -> pd.Series:
        protospacer_match_suffix = "_ProtospacerMatch"
        surrogate_match_suffix = "_SurrogateMatch"
        whitelist_guide_reporter_df_product = whitelist_guide_reporter_df.merge(whitelist_guide_reporter_df, how='cross', suffixes=(protospacer_match_suffix, surrogate_match_suffix))
        counterseries: pd.Series = whitelist_guide_reporter_df_product.apply(lambda pairwise_reporter: counterdict[tuple(pairwise_reporter[[index for index in pairwise_reporter.index if index.endswith(protospacer_match_suffix)]]), tuple(pairwise_reporter[[index for index in pairwise_reporter.index if index.endswith(surrogate_match_suffix)]])], axis=1)
        counterseries.index = pd.MultiIndex.from_frame(whitelist_guide_reporter_df_product)
        return counterseries
    

    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results = SurrogateProtospacerMismatchSetWhitelistReporterCounterSeriesResults()
    
    # MATCH counters
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_noncollapsed_match_counterseries = ambiguous_ignored_umi_noncollapsed_match_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_collapsed_match_counterseries = ambiguous_ignored_umi_collapsed_match_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_match_counterseries = ambiguous_ignored_match_counterdict

    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_noncollapsed_match_counterseries = ambiguous_accepted_umi_noncollapsed_match_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_collapsed_match_counterseries  = ambiguous_accepted_umi_collapsed_match_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_match_counterseries = ambiguous_accepted_match_counterdict

    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_noncollapsed_match_counterseries = ambiguous_spread_umi_noncollapsed_match_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_collapsed_match_counterseries = ambiguous_spread_umi_collapsed_match_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_match_counterseries = ambiguous_spread_match_counterdict

    # MISMATCH counters (keys are PAIRS of indices, representing the protospacer and surrogate match separately)
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_noncollapsed_mismatch_counterseries = ambiguous_ignored_umi_noncollapsed_mismatch_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_umi_collapsed_mismatch_counterseries = ambiguous_ignored_umi_collapsed_mismatch_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_ignored_mismatch_counterseries = ambiguous_ignored_mismatch_counterdict

    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_noncollapsed_mismatch_counterseries = ambiguous_accepted_umi_noncollapsed_mismatch_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_umi_collapsed_mismatch_counterseries = ambiguous_accepted_umi_collapsed_mismatch_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_accepted_mismatch_counterseries = ambiguous_accepted_mismatch_counterdict

    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_noncollapsed_mismatch_counterseries = ambiguous_spread_umi_noncollapsed_mismatch_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_umi_collapsed_mismatch_counterseries = ambiguous_spread_umi_collapsed_mismatch_counterdict
    surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results.ambiguous_spread_mismatch_counterseries = ambiguous_spread_mismatch_counterdict

    return surrogate_protospacer_mismatch_set_whitelist_reporter_counter_series_results


#
# CALLED FUNCTION TO RETRIEVE ALL MATCHSET AND MISMATCHSET COUNTERSERIES
#
@typechecked
def get_counterseries_all_results(observed_guide_reporter_umi_counts_inferred: GeneralMappingInferenceDict, whitelist_guide_reporter_df: pd.DataFrame, contains_barcode: bool, contains_surrogate: bool,  contains_umi: bool) -> AllMatchSetWhitelistReporterCounterSeriesResults:
    all_match_set_whitelist_reporter_counter_series_results = AllMatchSetWhitelistReporterCounterSeriesResults()
    all_match_set_whitelist_reporter_counter_series_results.protospacer_match = get_matchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_umi, "protospacer_match")
    if contains_barcode:
        all_match_set_whitelist_reporter_counter_series_results.protospacer_match_barcode_match = get_matchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_umi, "protospacer_match_barcode_match")
        if contains_surrogate:
            all_match_set_whitelist_reporter_counter_series_results.protospacer_match_surrogate_match_barcode_match = get_matchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_umi, "protospacer_match_surrogate_match_barcode_match")
            
            all_match_set_whitelist_reporter_counter_series_results.protospacer_mismatch_surrogate_match_barcode_match = get_mismatchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_umi, "protospacer_mismatch_surrogate_match_barcode_match")
    if contains_surrogate:
        all_match_set_whitelist_reporter_counter_series_results.protospacer_match_surrogate_match = get_matchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_umi, "protospacer_match_surrogate_match")
        all_match_set_whitelist_reporter_counter_series_results.protospacer_mismatch_surrogate_match = get_mismatchset_counterseries(observed_guide_reporter_umi_counts_inferred, whitelist_guide_reporter_df, contains_umi, "protospacer_mismatch_surrogate_match")
    
    return all_match_set_whitelist_reporter_counter_series_results



import csv
import multiprocessing as mp
from collections import Counter
from functools import reduce
from typeguard import typechecked
from functools import partial
from typing import Union, Tuple
from typing import Counter as CounterType

# DEPRRECATED CLASS

def gen_chunks(reader, chunksize=1000):
    """
    Chunk generator. Take a CSV `reader` and yield
    `chunksize` sized slices.
    Source: https://gist.github.com/miku/820490
    """
    chunk = []
    for index, line in enumerate(reader):
        if index % chunksize == 0 and index > 0:
            yield chunk
            chunk = []
        chunk.append(line)
    yield chunk

def process(tsv_chunk, include_surrogate = False):
    local_counter = Counter()
    
    for row in tsv_chunk.copy():
        if include_surrogate:
            local_counter[(row[1], row[2], row[3])] += 1
        else:
            local_counter[row[1]] += 1
    
    return local_counter


@typechecked
def get_reporter_tsv_observed_sequence_counts(reporter_tsv_fn:str, include_surrogate = False, cores=1) -> Union[CounterType[Tuple[str,str,str]], CounterType[str]]: 
    with open(reporter_tsv_fn, "r", newline='') as reporter_tsv_handler:
        tsv_reader = csv.reader(reporter_tsv_handler, delimiter='\t')  # change delimiter for normal csv files
        header = next(tsv_reader)
        
        read_chunks = gen_chunks(tsv_reader, chunksize=10000)

        process_func = partial(process, include_surrogate=include_surrogate)

        combined_counter = None
        if cores > 1:
            with mp.Pool(processes=cores) as pool:
                chunk_counters = pool.map(process_func, read_chunks)
                combined_counter = reduce(lambda x, y: x + y, chunk_counters)
        else:
            chunk_counters = [process_func(read_chunk) for read_chunk in read_chunks]
            combined_counter = reduce(lambda x, y: x + y, chunk_counters)

        return combined_counter 





###
### GUIDE PARSING HELPER FUNCTIONS
###
from typeguard import typechecked
from typing import Union, Optional, List, Tuple, DefaultDict
from typing import Counter as CounterType
from ..models.mapping_models import ProtospacerCounter, ProtospacerDictUMICounter, ProtospacerBarcodeCounter, ProtospacerBarcodeDictUMICounter, ProtospacerSurrogateCounter, ProtospacerSurrogateDictUMICounter, ProtospacerSurrogateBarcodeCounter, ProtospacerSurrogateBarcodeDictUMICounter
from collections import Counter
from typing import Callable
from functools import partial
from datetime import datetime
import gzip
import re
from Bio.Seq import Seq
from collections import Counter, defaultdict

#
# CLASS DEPRECATED AS OF 11/21/2024, use reporter_standard_fastq_parsing class
#


# This is for grouping the FASTQ lines in 1 tuple.
def grouper(iterable, n=4):
    "s -> (s0,s1,...sn-1), (sn,sn+1,...s2n-1), (s2n,s2n+1,...s3n-1), ..."
    return zip(*[iter(iterable)]*n)

# NOTE: Unsure if this function is used anymore
def perform_sequence_count(sequence_counter, r1_protospacer_fastq_filehandler, r2_surrogate_fastq_filehandler, barcode_pattern_regex: Optional[str] = None, umi_pattern_regex: Optional[str] = None):
    if r2_surrogate_fastq_filehandler is None:
        fastq_read_grouper = grouper(r1_protospacer_fastq_filehandler)   
    else:
        fastq_read_grouper = grouper(zip(r1_protospacer_fastq_filehandler, r2_surrogate_fastq_filehandler))

    for fastq_paired_read_group in fastq_read_grouper:
        if r2_surrogate_fastq_filehandler is None:
            r1_header_read = fastq_paired_read_group[0]
            r1_sequence_read = fastq_paired_read_group[1]
            protospacer=r1_sequence_read
            if barcode_pattern_regex:
                barcode_match = re.search(barcode_pattern_regex, r1_header_read)
                barcode = barcode_match.group(1)
                if umi_pattern_regex:
                    umi_match = re.search(umi_pattern_regex, r1_header_read)        
                    umi = umi_match.group(1)
                else:
                    pass
            else:
                if umi_pattern_regex:
                    umi = re.search(umi_pattern_regex, r1_header_read)        
                else:
                    pass
        else:
            r1_header_read, _ = fastq_paired_read_group[0]
            r1_sequence_read, r2_sequence_read = fastq_paired_read_group[1]
            surrogate=r1_sequence_read
        
        

        if barcode_pattern_regex:
            barcode = re.search(barcode_pattern_regex, r1_header_read)
        if umi_pattern_regex:
            umi = re.search(umi_pattern_regex, r1_header_read)
        

        sequence_counter[(protospacer, surrogate, barcode)][umi] += 1

    return sequence_counter

@typechecked
def get_umitools_observed_sequence_counts(r1_protospacer_fastq_file: str, r2_surrogate_fastq_file: Optional[str] = None, barcode_pattern_regex: Optional[str] = None, umi_pattern_regex: Optional[str] = None, revcomp_protospacer: bool = False, revcomp_surrogate: bool = True, revcomp_barcode: bool = True) -> Union[CounterType[str], DefaultDict[str, CounterType[str]], CounterType[Tuple[str, str]], DefaultDict[Tuple[str, str], CounterType[str]], CounterType[Tuple[str, str, str]], DefaultDict[Tuple[str, str, str], CounterType[str]]]:
    barcode_pattern_regex = None if ((barcode_pattern_regex is not None) and  (barcode_pattern_regex.strip() == "")) else barcode_pattern_regex
    umi_pattern_regex = None if ((umi_pattern_regex is not None) and (umi_pattern_regex.strip() == "")) else umi_pattern_regex

    print(f"Barcode pattern: {barcode_pattern_regex}")
    print(f"UMI pattern: {umi_pattern_regex}")

    revcomp = lambda sequence, do_revcomp: str(Seq(sequence).reverse_complement()) if do_revcomp else sequence

    def parse_fastq(r1_protospacer_fastq_filehandler, r2_surrogate_fastq_filehandler):
        if r2_surrogate_fastq_file is None: # ONLY R1
            fastq_single_read_grouper = grouper(r1_protospacer_fastq_filehandler)
            if barcode_pattern_regex is None: # ONLY R1; NO BARCODE
                if umi_pattern_regex is None: # ONLY R1; NO BARCODE; NO UMI
                    sequence_counter: ProtospacerCounter = Counter()
                    for fastq_single_read_group in fastq_single_read_grouper:
                        r1_sequence_read = fastq_single_read_group[1]
                        protospacer=revcomp(r1_sequence_read.strip(), revcomp_protospacer)
                        sequence_counter[protospacer] += 1
                else: # ONLY R1; NO BARCODE; YES UMI
                    sequence_counter: ProtospacerDictUMICounter = defaultdict(Counter)
                    for fastq_single_read_group in fastq_single_read_grouper:
                        r1_header_read = fastq_single_read_group[0]
                        r1_sequence_read = fastq_single_read_group[1]
                        protospacer=revcomp(r1_sequence_read.strip(), revcomp_protospacer)
                        umi = re.search(umi_pattern_regex, r1_header_read).group(1).strip()
                        sequence_counter[protospacer][umi] += 1
            else: # ONLY R1; YES BARCODE
                if umi_pattern_regex is None: # ONLY R1; YES BARCODE; NO UMI
                    sequence_counter: ProtospacerBarcodeCounter = Counter()
                    for fastq_single_read_group in fastq_single_read_grouper:
                        r1_sequence_read = fastq_single_read_group[1]
                        protospacer=revcomp(r1_sequence_read.strip(), revcomp_protospacer)
                        barcode = revcomp(re.search(barcode_pattern_regex, r1_header_read).group(1).strip(), revcomp_barcode)
                        sequence_counter[(protospacer, barcode)] += 1
                else: # ONLY R1; YES BARCODE; YES UMI
                    sequence_counter: ProtospacerBarcodeDictUMICounter = defaultdict(Counter)
                    for fastq_single_read_group in fastq_single_read_grouper:
                        r1_header_read = fastq_single_read_group[0]
                        r1_sequence_read = fastq_single_read_group[1]
                        protospacer=revcomp(r1_sequence_read.strip(), revcomp_protospacer)
                        barcode = revcomp(re.search(barcode_pattern_regex, r1_header_read).group(1).strip(), revcomp_barcode)
                        umi = re.search(umi_pattern_regex, r1_header_read).group(1).strip()
                        sequence_counter[(barcode, protospacer)][umi] += 1
        else: # YES R2
            fastq_paired_read_grouper = grouper(zip(r1_protospacer_fastq_filehandler, r2_surrogate_fastq_filehandler))
            if barcode_pattern_regex is None: # YES R2; NO BARCODE
                if umi_pattern_regex is None: # YES R2; NO BARCODE; NO UMI
                    sequence_counter: ProtospacerSurrogateCounter = Counter()
                    for fastq_paired_read_group in fastq_paired_read_grouper:
                        r1_sequence_read, r2_sequence_read = fastq_paired_read_group[1]
                        protospacer=revcomp(r1_sequence_read.strip(), revcomp_protospacer)
                        surrogate=revcomp(r2_sequence_read.strip(), revcomp_surrogate)
                        sequence_counter[(protospacer, surrogate)] += 1
                else: # YES R2; NO BARCODE; YES UMI
                    sequence_counter: ProtospacerSurrogateDictUMICounter = defaultdict(Counter)
                    for fastq_paired_read_group in fastq_paired_read_grouper:
                        r1_header_read, _ = fastq_paired_read_group[0]
                        r1_sequence_read, r2_sequence_read = fastq_paired_read_group[1]
                        
                        protospacer=revcomp(r1_sequence_read.strip(), revcomp_protospacer)
                        surrogate=revcomp(r2_sequence_read.strip(), revcomp_surrogate)
                        umi = re.search(umi_pattern_regex, r1_header_read).group(1).strip()
                        
                        sequence_counter[(protospacer, surrogate)][umi] += 1
            else: # YES R2; YES BARCODE
                if umi_pattern_regex is None: # YES R2; YES BARCODE; NO UMI
                    sequence_counter: ProtospacerSurrogateBarcodeCounter = Counter()
                    for fastq_paired_read_group in fastq_paired_read_grouper:
                        r1_header_read, _ = fastq_paired_read_group[0]
                        r1_sequence_read, r2_sequence_read = fastq_paired_read_group[1]
                        
                        protospacer=revcomp(r1_sequence_read.strip(), revcomp_protospacer)
                        surrogate=revcomp(r2_sequence_read.strip(), revcomp_surrogate)
                        barcode = revcomp(re.search(barcode_pattern_regex, r1_header_read).group(1).strip(), revcomp_barcode)
                        sequence_counter[(protospacer, surrogate, barcode)] += 1
                else: # YES R2; YES BARCODE; YES UMI
                    sequence_counter: ProtospacerSurrogateBarcodeDictUMICounter = defaultdict(Counter)
                    for fastq_paired_read_group in fastq_paired_read_grouper:
                        r1_header_read, _ = fastq_paired_read_group[0]
                        r1_sequence_read, r2_sequence_read = fastq_paired_read_group[1]
                        
                        protospacer=revcomp(r1_sequence_read.strip(), revcomp_protospacer)
                        surrogate=revcomp(r2_sequence_read.strip(), revcomp_surrogate)
                        barcode = revcomp(re.search(barcode_pattern_regex, r1_header_read).group(1).strip(), revcomp_barcode)
                        umi = re.search(umi_pattern_regex, r1_header_read).group(1).strip()
                        
                        sequence_counter[(protospacer, surrogate, barcode)][umi] += 1
        return sequence_counter
    
    before_file_loading_time = datetime.now()
    # Open R1 file handler
    r1_protospacer_fastq_filehandler = None
    if r1_protospacer_fastq_file.endswith('.gz'):
        print(f"Opening FASTQ.gz file with gzip, filename={r1_protospacer_fastq_file}")
        r1_protospacer_fastq_filehandler = gzip.open(r1_protospacer_fastq_file, "rt", encoding="utf-8")
    else:
        print(f"Opening FASTQ file, filename={r1_protospacer_fastq_file}")
        r1_protospacer_fastq_filehandler = open(r1_protospacer_fastq_file, "r")
    
    # Open R2 file handler if provided
    r2_surrogate_fastq_filehandler = None
    if r2_surrogate_fastq_file is not None:
        if r2_surrogate_fastq_file.endswith('.gz'):
            print(f"Opening FASTQ.gz file with gzip, filename={r2_surrogate_fastq_file}")
            r2_surrogate_fastq_filehandler = gzip.open(r2_surrogate_fastq_file, "rt", encoding="utf-8")
        else:
            print(f"Opening FASTQ file, filename={r2_surrogate_fastq_file}")
            r2_surrogate_fastq_filehandler = open(r2_surrogate_fastq_file, "r")
    after_file_loading_time = datetime.now()
    print(f"{(after_file_loading_time-before_file_loading_time).seconds} seconds for file loading")

    sequence_counter = parse_fastq(r1_protospacer_fastq_filehandler, r2_surrogate_fastq_filehandler)
    after_parsing_time = datetime.now()
    print(f"{(after_parsing_time-after_file_loading_time).seconds} seconds for parsing")

    # Close the file handlers when done
    if r1_protospacer_fastq_filehandler is not None:
        r1_protospacer_fastq_filehandler.close()
    if r2_surrogate_fastq_filehandler is not None:
        r2_surrogate_fastq_filehandler.close()

    return sequence_counter



from enum import Enum
from typing import Optional, Any
from dataclasses import dataclass
import pandas as pd


class GuideCountErrorType(Enum):
    NO_MATCH = "No guide found within hamming distance"
    MULTIPLE_MATCH = "Multiple exact matches found for guide (likely a truncated guide read assuming guide series is unique)"
    NO_PROTOSPACER_WITH_SAME_LENGTH = "No whitelisted guides with same length as observed guide - maybe try enabling truncating whitelisted guides"
    NO_BARCODE_WITH_SAME_LENGTH = "No whitelisted barcode with the same length as the observed"
    NO_SURROGATE_WITH_SAME_LENGTH = "No whitelisted surrogate with the same length as the observed"
    NO_MATCH_PROTOSPACER_HAMMING_THRESHOLD = "No whitelisted guide that is below the provided guide hamming distance threshold"
    NO_MATCH_BARCODE_HAMMING_THRESHOLD = "No whitelisted barcode that is below the provided barcode hamming distance threshold"
    NO_MATCH_SURROGATE_HAMMING_THRESHOLD = "The inferred whitelisted surrogate does not match with the observed surrogate below the given hamming distance threshold"
    MULTIPLE_MATCH_EXACT = "Multiple exact guide matches, double check that there are no duplicates in your guide library (especially if truncation is enabled)"
    NO_PROTOSPACER_MATCH_MISSING_INFO = "The protospacer was not provided"
    NO_SURROGATE_MATCH_MISSING_INFO = "The surrogate was not provided"
    NO_BARCODE_MATCH_MISSING_INFO = "The barcode was not provided"
    NO_MATCH_OBSERVED_SURROGATE_SHORTER = "The observed surrogate is shorter than the inferred surrogate"
    NULL_MATCH_RESULT = "Match result was null, likely since match result type is not supported given the inputs (i.e. a barcode_match result when barcode is not specified)"

@dataclass
class GuideCountError:
    guide_count_error_type: GuideCountErrorType
    miscellaneous_info_dict: dict = None

@dataclass
class InsufficientLengthGuideCountError(GuideCountError):
    # NOTE: Added default argument to prevent error of having non-default after defaults in child classes
    sequence_length: Optional[int]  = None
    minimum_length: Optional[int] = None

@dataclass
class ProtospacerInsufficientLengthGuideCountError(InsufficientLengthGuideCountError):
    guide_count_error_type: GuideCountErrorType = GuideCountErrorType.NO_PROTOSPACER_WITH_SAME_LENGTH

@dataclass
class SurrogateInsufficientLengthGuideCountError(InsufficientLengthGuideCountError):
    guide_count_error_type: GuideCountErrorType = GuideCountErrorType.NO_SURROGATE_WITH_SAME_LENGTH

@dataclass
class BarcodeInsufficientLengthGuideCountError(InsufficientLengthGuideCountError):
    guide_count_error_type: GuideCountErrorType = GuideCountErrorType.NO_BARCODE_WITH_SAME_LENGTH

@dataclass
class MissingInfoGuideCountError(GuideCountError):
    sequence_value: Optional[Any] = None

@dataclass
class ProtospacerMissingInfoGuideCountError(MissingInfoGuideCountError):
    guide_count_error_type: GuideCountErrorType = GuideCountErrorType.NO_PROTOSPACER_MATCH_MISSING_INFO

@dataclass
class SurrogateMissingInfoGuideCountError(MissingInfoGuideCountError):
    guide_count_error_type: GuideCountErrorType = GuideCountErrorType.NO_SURROGATE_MATCH_MISSING_INFO

@dataclass
class BarcodeMissingInfoGuideCountError(MissingInfoGuideCountError):
    guide_count_error_type: GuideCountErrorType = GuideCountErrorType.NO_BARCODE_MATCH_MISSING_INFO

@dataclass
class HammingThresholdGuideCountError(GuideCountError):
    hamming_min: Optional[int] = None
    hamming_threshold: Optional[int] = None
    original_df: Optional[pd.DataFrame] = None
    hamming_min_match_df: Optional[pd.DataFrame] = None

@dataclass
class ProtospacerHammingThresholdGuideCountError(HammingThresholdGuideCountError):
    guide_count_error_type: GuideCountErrorType = GuideCountErrorType.NO_MATCH_PROTOSPACER_HAMMING_THRESHOLD
    barcode_subsetted: Optional[bool] = None

@dataclass
class SurrogateHammingThresholdGuideCountError(HammingThresholdGuideCountError):
    guide_count_error_type: GuideCountErrorType = GuideCountErrorType.NO_MATCH_SURROGATE_HAMMING_THRESHOLD
    barcode_subsetted: Optional[bool] = None

@dataclass
class BarcodeHammingThresholdGuideCountError(HammingThresholdGuideCountError):
    guide_count_error_type: GuideCountErrorType = GuideCountErrorType.NO_MATCH_BARCODE_HAMMING_THRESHOLD